// Code generated by cmd/cgo -godefs; DO NOT EDIT.
// cgo -godefs defs_openbsd.go

package route

const (
	sizeofRtMsghdr = 0x60

	sizeofSockaddrStorage = 0x100
	sizeofSockaddrInet    = 0x10
	sizeofSockaddrInet6   = 0x1c
)
