// Copyright 2012 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package cgotest

import "testing"

// Stubs for tests that fails to build on Android
func test6997(t *testing.T)        {}
func test8694(t *testing.T)        {}
func testSigaltstack(t *testing.T) {}
