// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

#include "textflag.h"

TEXT ·set_crosscall2(SB),NOSPLIT,$0-0
	UNDEF

TEXT crosscall2(SB), NOSPLIT, $0
	UNDEF
