// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This file silences errors about body-less functions
// that are provided by intrinsics in the latest version of the compiler,
// but may not be known to the bootstrap compiler.
