// Code generated by running "go generate" in golang.org/x/text. DO NOT EDIT.

// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build !go1.18

package idna

const transitionalLookup = true
